#include <iostream>
#include <thread>
#include <condition_variable>
#include <mutex>
#include <unistd.h>
#include <cstring>
#include <RtAudio.h>
#include <atomic>
#include "ringbuffer.h"

#include "osc/OscReceivedElements.h"
#include "osc/OscPacketListener.h"
#include "oscpack_1_1_0/ip/UdpSocket.h"
#include "osc/OscOutboundPacketStream.h"


#define PORT 3000
#define SENDPORT 3002
#define OUTPUT_BUFFER_SIZE 1024
#define ADDRESS "169.231.126.143"
#define SINEBUFFERSIZE 4096
#define SAMPLERATE 44100

#define PHSMASK64 0x0FFFFFFFFFFFFFFFL
#define MAXLEN64  0x1000000000000000L

using namespace std;

typedef struct
{
    int runWriter;
    RingBuffer<double> *rb;
    double writeArray[2048];
    double readArray[2048];
    condition_variable cv;
    mutex m;

    //wavetable
    double sineBuffer[SINEBUFFERSIZE];

    //osc 1
    double freq;
    uint64_t phs;
    uint64_t inc;

    //osc 2
    double freq2;
    uint64_t phs2;
    uint64_t inc2;

    //osc 3
    double freq3;
    uint64_t phs3;
    uint64_t inc3;
    double modDepth;

    int sr;
    uint64_t lowbits;

    RingBuffer<double> *rbAmp;
    RingBuffer<double> *rbFreq;
    RingBuffer<double> *rbFreq2;
    RingBuffer<double> *rbFreq3;

    atomic<int> osc1Enabled;
    atomic<int> osc2Enabled;
    atomic<int> osc3Enabled;

    double amp;

} rtUserData;

rtUserData userData;


class ExamplePacketListener : public osc::OscPacketListener {
protected:

    virtual void ProcessMessage( const osc::ReceivedMessage& m,
                                 const IpEndpointName& remoteEndpoint )
    {
        (void) remoteEndpoint; // suppress unused parameter warning

        try{

            if( std::strcmp( m.AddressPattern(), "/slider" ) == 0 ){
                osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
                int sliderNum;
                float value;
                args >> sliderNum >> value >> osc::EndMessage;

                if(sliderNum==0){
                    userData.rbAmp->write(value);
                }else if(sliderNum == 1){
                    userData.rbFreq->write(value*1000);
                }else if(sliderNum == 2){
                    userData.rbFreq2->write(value*1000);
                }else if(sliderNum == 3){
                    userData.rbFreq3->write(value*1000);
                }

            }else if(std::strcmp( m.AddressPattern(), "/orientation" ) == 0 ){
                osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
                float x,y,z;
                args >> x >> y >> z >> osc::EndMessage;
                float scaledX = (((x-(-M_PI))*1000.0)/ M_PI) + 0.0;
                float scaledY = (((y-(-M_PI))*1000.0)/ M_PI) + 0.0;
                float scaledZ = (((z-(-M_PI))*1000.0)/ M_PI) + 0.0;
                userData.rbFreq->write(scaledX);
                userData.rbFreq2->write(scaledY);
                userData.rbFreq3->write(scaledZ);


            }else if(std::strcmp( m.AddressPattern(), "/switch" ) == 0 ){
                osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
                int index,value;
                args >> index >>value  >>osc::EndMessage;
                if(index==1){
                    userData.osc1Enabled = value;
                }else if(index == 2){
                    userData.osc2Enabled = value;
                }else if(index == 3){
                    userData.osc3Enabled = value;
                }

            }


            else if( std::strcmp( m.AddressPattern(), "/test2" ) == 0 ){
                osc::ReceivedMessage::const_iterator arg = m.ArgumentsBegin();
                bool a1 = (arg++)->AsBool();
                int a2 = (arg++)->AsInt32();
                float a3 = (arg++)->AsFloat();
                const char *a4 = (arg++)->AsString();
                if( arg != m.ArgumentsEnd() )
                    throw osc::ExcessArgumentException();

                std::cout << "received '/test2' message with arguments: "
                          << a1 << " " << a2 << " " << a3 << " " << a4 << "\n";
            }
        }catch( osc::Exception& e ){
            // any parsing errors such as unexpected argument types, or
            // missing arguments get thrown as exceptions.
            std::cout << "error while parsing message: "
                      << m.AddressPattern() << ": " << e.what() << "\n";
        }
    }
};


UdpTransmitSocket transmitSocket( IpEndpointName( ADDRESS, SENDPORT ) );
char buffer[OUTPUT_BUFFER_SIZE];

void writeThread()
{
    while(userData.runWriter) {
        unique_lock<mutex> lk(userData.m);
        userData.cv.wait(lk);
        double rmsAccumL = 0;
        double rmsAccumR = 0;
        while (userData.rb->read(userData.readArray, 256* 2) == 512) {
            for(int i = 0; i < 256; i++){
                rmsAccumL += pow(userData.readArray[i*2],2.0);
                rmsAccumR += pow(userData.readArray[i*2 + 1],2.0);
            }

            float rmsL = sqrt((1.0/256.0)*rmsAccumL);
            float rmsR = sqrt((1.0/256.0)*rmsAccumR);
            osc::OutboundPacketStream p( buffer, OUTPUT_BUFFER_SIZE );
            p <<osc::BeginMessage( "/RMS" )<< rmsL << rmsR << osc::EndMessage;
            transmitSocket.Send( p.Data(), p.Size() );
        }
        lk.unlock();
    }
}

int audioCB( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
             double streamTime, RtAudioStreamStatus status, void *userData )
{
    (void)inputBuffer;
    (void)streamTime;
    unsigned int i, j;
    double *outbuffer = (double *) outputBuffer;
    rtUserData *data = (rtUserData*)userData;
    if ( status )
        std::cout << "Stream underflow detected!" << std::endl;


    data->rbAmp->read(data->amp);
    data->rbFreq->read(data->freq);
    data->inc = (u_int64_t) lrint(data->freq* (double) MAXLEN64/SAMPLERATE);

    data->rbFreq2->read(data->freq2);
    data->inc2 = (u_int64_t) lrint(data->freq2* (double) MAXLEN64/SAMPLERATE);
    data->rbFreq3->read(data->freq3);
    data->inc3 = (u_int64_t) lrint(data->freq3* (double) MAXLEN64/SAMPLERATE);

    for (i = 0;i<nBufferFrames; i++){
        for(j=0;j<2;j++){
            double out = 0.0;
            out +=data->sineBuffer[data->phs >> data->lowbits]*data->osc1Enabled;
            out += data->sineBuffer[data->phs2 >> data->lowbits]*data->osc2Enabled;
            out += data->sineBuffer[data->phs3 >> data->lowbits]*data->osc3Enabled;
            out *= 0.2*data->amp;
            *outbuffer++= out;
            data->writeArray[i*2 + j] = out;
        }

        data->phs = (data->phs + data->inc)& PHSMASK64;
        data->phs2 = (data->phs2 + data->inc2)& PHSMASK64;
        data->phs3 = (data->phs3 + data->inc3)& PHSMASK64;
    }

    int numWritten = data->rb->write(data->writeArray, nBufferFrames * 2);
    if (numWritten == 0) {
        printf("Couldn't write to ring buffer.");
    } else {
        data->cv.notify_all();
    }

    return 0;
}



int main()
{
    RtAudio audio;
    unsigned int sampleRate = SAMPLERATE;
    unsigned int bufferFrames = 256; // 256 sample frames


    userData.amp = 0.7;
    userData.rb = new RingBuffer<double>(bufferFrames * 2 * 8);
    userData.rbAmp = new RingBuffer<double>(bufferFrames * 2 * 8);
    userData.rbFreq = new RingBuffer<double>(bufferFrames * 2 * 8);
    userData.rbFreq2 = new RingBuffer<double>(bufferFrames * 2 * 8);
    userData.rbFreq3 = new RingBuffer<double>(bufferFrames * 2 * 8);


    for(int i = 0; i < SINEBUFFERSIZE;i++){
        userData.sineBuffer[i] = sin(2.0 * M_PI * i / (double) SINEBUFFERSIZE);
    }

    userData.osc1Enabled = 1;
    userData.osc2Enabled = 1;
    userData.osc3Enabled = 1;
    userData.freq = 440.0;
    userData.phs = 0;
    userData.inc = (u_int64_t) lrint(userData.freq* (double) MAXLEN64/SAMPLERATE);
    userData.freq2 = 480.0;
    userData.phs2 = 0;
    userData.inc2 = (u_int64_t) lrint(userData.freq2* (double) MAXLEN64/SAMPLERATE);


    userData.freq3 = 500.0;
    userData.phs3 = 0;
    userData.inc3 = (u_int64_t) lrint(userData.freq3* (double) MAXLEN64/SAMPLERATE);;


    userData.sr = SAMPLERATE;
    userData.lowbits = log2(MAXLEN64/SINEBUFFERSIZE);


    userData.runWriter = 1;
    thread sfWriter(writeThread);


    ExamplePacketListener listener;
    UdpListeningReceiveSocket s(
                IpEndpointName( IpEndpointName::ANY_ADDRESS, PORT ),
                &listener );


    RtAudio::StreamParameters parameters;
    parameters.deviceId = audio.getDefaultOutputDevice();
    parameters.nChannels = 2;
    parameters.firstChannel = 0;
    RtAudio::StreamParameters inparameters;
    inparameters.deviceId = audio.getDefaultInputDevice();
    inparameters.nChannels = 2;
    inparameters.firstChannel = 0;

    try {
        audio.openStream( &parameters, &inparameters, RTAUDIO_FLOAT64,
                          sampleRate, &bufferFrames, &audioCB, &userData );
        audio.startStream();
    }
    catch (RtAudioError &error) {
        // Handle the exception here
        error.printMessage();
        return -1;
    }

    std::cout << "\npress <ctrl+c> to quit.\n";

    s.RunUntilSigInt();
    //std::cin.get(input);


    try {
        // Stop and close the stream
        audio.stopStream();
        audio.closeStream();
    }
    catch (RtAudioError &error) {
        error.printMessage();
    }


    userData.runWriter = 0;
    userData.cv.notify_all();
    sfWriter.join();

    cout << "Program Exited" << endl;
    return 0;
}

