#include "ringbuffer.h"
#include <memory>
#include <cstring>

template<typename T> RingBuffer<T>::RingBuffer(int size) :
    m_size(size)
{
    m_writeptr = m_readptr = 0;
    m_buffer = (T *) calloc(size, sizeof(T));
}

template<typename T> RingBuffer<T>::~RingBuffer()
{
    free(m_buffer);
}

template<typename T>
int RingBuffer<T>::availableRead()
{
    // Could use if statements which may be faster
    return (m_size + m_writeptr - m_readptr) % m_size;
}

template<typename T>
int RingBuffer<T>::availableWrite()
{
    return (m_size + m_readptr - m_writeptr - 1) % m_size;
}

template<typename T>
int RingBuffer<T>::read(T *data, int numElements)
{
    int available = availableRead();
    int vectorSize = numElements;
    int copySize = available;
    if(available >= vectorSize) {
        copySize = available < vectorSize ? available : vectorSize;
        if (m_readptr < m_writeptr || (m_size - m_readptr) > vectorSize) {
            memcpy(data, m_buffer + m_readptr, copySize * sizeof(T));
        } else {
            memcpy(data, m_buffer + m_readptr, (m_size - m_readptr) * sizeof(T));
            memcpy(data, m_buffer + (m_size - m_readptr), m_writeptr * sizeof(T));
        }
        atomic_fetch_add(&m_readptr, copySize);
        if (m_readptr >= m_size) {
            atomic_fetch_sub(&m_readptr, m_size);
        }
    } else {
        return 0;
    }
    return copySize;
}

template<typename T>
int RingBuffer<T>::read(T &value)
{
    if(availableRead()) {
        value = m_buffer[m_readptr];
        atomic_fetch_add(&m_readptr, 1);
        return 1;
    }
    return 0;
}

template<typename T>
int RingBuffer<T>::write(T *data, int numElements)
{
    int available = availableWrite();
    int vectorSize = numElements;
    int copySize = available;
    if(available >= vectorSize) {
        copySize = available < vectorSize ? available : vectorSize;
        if (m_writeptr <= m_readptr || (m_size - m_writeptr) > vectorSize) {
            memcpy(m_buffer + m_writeptr, data, copySize * sizeof(T));
        } else {
            memcpy(m_buffer + m_writeptr, data, (m_size - m_writeptr) * sizeof(T));
            memcpy(m_buffer, data + (m_size - m_writeptr), m_readptr * sizeof(T));
        }
        atomic_fetch_add(&m_writeptr, copySize);
        if (m_writeptr >= m_size) {
            atomic_fetch_sub(&m_writeptr, m_size);
        }
    } else {
        return 0;
    }
    return copySize;
}

template<typename T>
int RingBuffer<T>::write(T value)
{
    if (availableWrite()) {
        m_buffer[m_writeptr] = value;
        atomic_fetch_add(&m_writeptr, 1);
        return 1;
    }
    return 0;
}

